package CommSecure;
#
## This script is available under the Gnu General Public Licence (GPL)
##
## Written by James Bromberger (james@rcpt.to).
##
#
use strict;
my $VERSION="0.3";
sub Version { $VERSION; }

use CommSecure::Request;
use CommSecure::Validate;
use CommSecure::Cleared;

=head1 NAME

CommSecure - A comeplete fincancial transaction facility interface

=head1 SYNOPSIS

  use CommSecure;
  my $cs_t = CommSecure->new(-dbh => $dbh);

=head1 DESCRIPTION

This library implments a tracking mechanism that permits one to request, 
validate/authorise, and capture confirmation of financial transactions 
with CommSecure, a credit card processing broker.

This library, and the three modules that it depends upon, require that three 
tables are available in a database, and that it has SELECT and INSERT 
permissions on these tables. It does not require UPDATE, DELETE, or any 
other permissions; it probably should *not* have them for security reasons.

The main steps of the processing are:

=over 3

=item *

Start a new request for payment by creating a Request object 
(and presenting a request form to the web client via CGI)

=item * 

Validate or decline a request for verification by creating a Validate 
object( and returning Accept or Decline to the web client via CGI)

=item *

Accept confirmation (or rejection) of a transaction

=back

Each part of the transaction (request, validate, confirmation) requires a 
database table be present with some fields. We will record every attempt 
to create an object - however, we note that some attempts may be 
incorrect, misleading, or invalid. For exmaple, we may recieve a verification 
without (or before) a request is generated.

We identify each trasnaction by a Refernece Number (RefNum). This is a unique 
string.

Hence, a complete transaction may be Valid or Invalid, based upon the 
contstraints of order and dependency of processing. A transaction is said to 
be valid if:

=over 3

=item * 

It has a request, but no verify or confirmation

=item * 

It has a request and a verify, but no confirmation, and the time of 
the request is earlier than the verify, and the amount to transfer as 
mentioned in the verify request matches the amount in the request.

=item * 

It has a request, and a verify, and a confirmation, and the time of 
the request is earlier than the verify, and the time of the confirmation is 
earlier than the verify, and the amount of the confirmation matches the 
amount in the verify and matches the amoutn in the request.

=back

Any transaction that is not in this order is deemed Invalid.

Given this dependency, we note that if users start to process a 
transaction, they may at any time stop processing this by simply not 
completing the routine, which is beyond our control. Hence, we must 
recognise transactions that are no longer being processed. We assign 
some timeout to these transactions, and consider them Aborted after 
this time; however they may be valid.

Thus, after generating a request, no verification attempt is made inside one 
hour; we deem this transaction Aborted. If a verficiation is attempted after 
this period, we record it, but decline.

=head1 METHODS

 my @transactions = CommSecure->find(-dbh => $dbh);
 $cs_t->refnum;
 $cs_t->is_valid;
 $cs_t->is_complete;
 $cs_t->is_approved;
 $cs_t->is_declined;
 $cs_t->as_text;

 # The sub-objects
 $cs_t->request;
 $cs_t->validate;
 $cs_t->result;

 # Reporting
 print CommSecure->summary_html(-dbh => $dbh);

=head2 find

This method takes a handle to a database and returns an array of 
CommSecure objects based upon a search criteria. The options 
are:

=over 3

=item *

-dbh - A reference to a database handle (DBI::DBD)

=item *

-refnum - The exact reference number of a transaction. This should be unique; 
you should get an array with one object in it if this exists.

=item *

-date_request_min - The oldest date of a request for payment, as a ctime 
integer (as returned by time()).

=item * 

-date_request_max - The youngest date of a request

=item * 

-date_verify_min - The oldest date of a verify

=item *

-date_verify_max - The oldest date of a verify

=item * 

-date_confirm_min - The oldest date of a confirm

=item * 

-date_confirm_max - The youngest date of a confirm

=item *

-custnum - The customer number hash; all transactions for a particular person

=item *

-amount - The exact amount of a transaction in cents (eg $2.50 => 250)

=item *

-amount_min - The minimum amount of the transaction 

=item *

-amount_max - The maximum amount of the transaction

=item *

-valid_only - Do not return invalid transactions

=item * 

-invalid_only - Do not return valid transactions

=item * 

-complete_only - Do not return incomplete transactions

=item * 

-incomplete_only - Do not return complete transactions

=item *

-approved_only - Only return approved, complete, valid transactions

=item *

-declined_only - Only return declined, complete, valid transactions

=back

Obviously if we use some combinations of the above then we will get none 
returned; use wisely.

=head2 refnum

Every unique transaction has a reference number (refnum) that ties together 
a request for payment, a verify, and a result.

=head2 is_valid

A transactionis valid if it has a request, and optionally a validate, 
and optionally a result, and that the date of the request for payment being 
raised is before the date of the verify (if any), and the verify 
is before the date of the result (if any).

Thus, we can never have a result without prior verification, and we cannot 
have a verification without a request for payment having first been 
generated.

=head2 is_complete

A transaction is complete if it is valid (is_valid) and has a result. Thus, 
it also has a complete set of request, verify, and result. It may 
or may not have been an approved transaction, but it is finished.

=head2 is_approved

A transaction is approved if it is complete (is_complete) and the end result 
was a confirmation of transfer of funds -- ie, and approval.

=head2 summary_html

This returns the contents of a report formatted as HTML, suitable for 
printing into the body of an HTML document. It details the state of 
current processing, any invalid transactions, and the amount currently cleared.


=head2 start_new_transaction

This will create a new transaction, which should then be verified. This will 
fail if we have already used the supplied RefNum in a previous transaction.

=head2 validate_transaction

This is the second stage in a transaction; it will save the parameters of the 
transaction for processing regardless of wether it has already been validated, 
since it is generating a log. However, the result will be true if this is a 
valid transaction that has not yet been verified and the paramters are 
consistent with the request for payment.

Generally the wrapper script will return 'Accepted' for true and 'Decline' for 
false.

=head2 confirm_transaction

This is the final step in a transaction; it will always save the paramaters 
of the transaction as it is making a log. However, it will only return true 
if this is a valid transaction that has not yet been given a result but has 
been verified.

Generally the wrapper script will return 'Accepted' for true and 'Decline' for 
false.

=head1 Database Structure

The implmentation of the payment system requires that three objects be present: 
a table for requests, implemented in the CommSecure::Request object, a 
table for verifies, implemented in the CommSecure::Validate object, and 
a table for the end result confirmation or deline, as implmented in the 
CommSecure::Cleared object.

We document these three objects here to keep them together; I am assuming 
MySQL as the RDBMS you are using; adjust to taste, since you pass a 
reference to a database handle in to the object and it just uses that.

We note that when executing this code as the effective web server user 
(eg, from CGI) then we should probably only grant SELECT and INSERT prrivs 
on these tables; at no time should the web interface delete or alter 
records already in the database.

=head2 Requests

 create table CommSecure_Requests (ID int primary key auto_increment, Date datetime not null, Host varchar(255) not null, MerchantNum varchar(255) not null, RefNum varchar(255) not null, Amount int not null, CustNum varchar(255) not null, MemberNum varchar(255), MerchandData blob, Description blob);

We note that most of the fields are mandatory. We will not be able to save a 
request unless we have these fields present. 

=head2 Verifys

 create table CommSecure_Validate (ID int primary key auto_increment, Date datetime not null, Host varchar(255) not null, RefNum varchar(255) not null, Amount int not null);

All fields here are mandatory. If an enquiry is recieved and it does not 
have all of these fields then we cannot save it. We may get multiple enquiries 
for confirmation; that is OK, we should be able to confirm multiple times.


=head2 Results

 create table CommSecure::Cleared (ID int primary key auto_increment, Date datetime not null, Host varchar(255) not null, RefNum varchar(255) not null, Amount int not null, Response varchar(5) not null, AuthNo varchar(255), Audit varchar(255), ErrorMsg varchar(255), Sig text);

Some fields are mandatory. We should only get one result at the end of 
the transaction: any more is suspicious.

=head1 MORE DOCUMENTATION

Reference the CommSecure document "web_protocols", and the CommSecure web 
site at http://www.commsecure.com.au/.

=head1 LICENCE

CommSecure.pm - Perl CommSecure library implmentation.
Copyright (C) 2002 James Bromberger <james@rcpt.to>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=head1 COPYRIGHT

Copyright 2002 James Bromberger, james@rcpt.to.

=cut



sub new {
	my $proto = shift;
	my $class = ref($proto)||$proto;
	my $self = {};
	bless $self, $class;
	my %args = @_;
	my $ok = 1;
	$self->dbh($args{'-dbh'}) if defined $args{'-dbh'};
	$self->load_refnum($args{'-refnum'}) if defined $args{'-refnum'};
	if (defined $args{'-date'}) {
		if (not $self->date($args{'-date'})) {
			warn "date bad";
			$ok = 0;
		}
	}
	return $self if $ok;
}

sub dbh {
	my $self = shift;
	if (@_) {
		my $new_dbh = shift;
		if (ref $new_dbh !~ "DBI::db") {
			warn "Not a DBI:db handle: $new_dbh";
			return;
		}
		$self->{'dbh'} = $new_dbh;
	}
	return $self->{'dbh'};
}

sub refnum {
	my $self = shift;
	return $self->request->refnum if (defined($self->request)&& defined($self->request->refnum));
	return $self->validate->refnum if (defined($self->validate)&& defined($self->validate->refnum));
	return $self->result->refnum if (defined($self->result) && defined($self->result->refnum));
}

sub custnum {
	my $self = shift;
	return $self->request->custnum if (defined($self->request) && defined($self->request->custnum));
	return $self->validate->custnum if (defined($self->validate) && defined($self->validate->custnum));
	return $self->result->custnum if (defined($self->result) && defined($self->result->custnum));
}


sub amount {
	my $self = shift;
	#warn "Getting amount for refnum " . $self->refnum if defined($self->refnum);
	if (defined($self->result)) {
		foreach (@{$self->result}) {
			return $_->amount if (defined($_) && defined($_->amount));
		}
	}
	if (defined($self->validate) && defined($self->validate->[0]->amount)) {
		#warn "Using validate amount for refnum " . $self->refnum;
		return $self->validate->[0]->amount;
	}
	return $self->request->amount if (defined($self->request) && defined($self->request->amount));
}

sub as_text {
	my $self = shift;
	# Return a string listing the refnum, if there is one. If there is 
	# no refnum, then there is no transaction, valid or invalid
	return unless ($self->refnum);
	my $string;

	$string.= sprintf "RefNum: %s", $self->refnum;

	if ($self->request) {
		$string.= "Requested";
		$string.= sprintf " on %s", scalar localtime $self->request->date if $self->request->date;
		$string.= sprintf " from %s", $self->request->host if $self->request->host;
		$string.= sprintf " for merchant %s", $self->request->merchantnum if $self->request->merchantnum;
		$string.= sprintf " for amount \$%d", $self->request->amount if defined $self->request->amount;
		$string.= sprintf " for customer %s", $self->request->custnum if defined $self->request->custnum;
		$string.= sprintf " for %s", $self->request->description if defined $self->request->description;
	}

	if ($self->validate) {
		$string.= "Validated";
		$string.= sprintf " on %s", scalar localhost $self->validate->date if defined $self->validate->date;
		$string.= sprintf " from %s", $self->validate->host if defined $self->validate->host;
		$string.= sprintf " for %s", $self->validate->amount if defined $self->validate->amount;
	}

	if ($self->result) {
		$string.= "Result";
		$string.= sprintf " on %s", scalar localhost $self->result->date if defined $self->result->date;
		$string.= sprintf " from %s", $self->result->host if defined $self->result->host;
		$string.= sprintf " for %s", $self->result->amount if defined $self->result->amount;
		$string.= sprintf " result %s", $self->result->response if defined $self->result->response;
		$string.= sprintf " authno %s", $self->result->authno if defined $self->result->authno;
		$string.= sprintf " audit %s", $self->result->audit if defined $self->result->audit;
		$string.= sprintf " error \"%s\"", $self->result->errormsg if defined $self->result->errormsg;
	}
	return $string;
}

sub summary_html {
	my $self = shift;
	my %args = @_;
	if (not (defined($args{'-dbh'}) and ref($args{'-dbh'}) =~ "DBI::db")) {
		warn "No DBH";
		return;
	}
	my $dbh = $args{'-dbh'};
	my $html;
	# Whats the general information...
	my @all = CommSecure->find(-dbh => $dbh);
	my @incomplete = CommSecure->find(-dbh => $dbh, -valid_only => 1, -incomplete_only =>1);
	my @invalid = CommSecure->find(-dbh => $dbh, -invalid_only => 1);

	$html.= "<h2>Executive Summary</h2>\n";
	$html.= sprintf "<p>There are %d transactions in total, of which there are %d incomplete and %d invalid.</p>\n", $#all +1, $#incomplete +1, $#invalid +1;
	$html.= "<table>";
	#warn "Have " . ($#all+1) . " transactions";
	$html.= CommSecure->_valid_reqs_html(-transactions => \@all);
	$html.= CommSecure->_valid_reqs_pending_verify_html(-transactions => \@all);
	$html.= CommSecure->_time_last_type_html(-transactions => \@all, -type => 'request');
	$html.= CommSecure->_cnt_val_avrg_time_to_type_html(-transactions => \@all, -type => 'validate');
	$html.= CommSecure->_time_last_type_html(-transactions => \@all, -type => 'validate');
	$html.= CommSecure->_cnt_val_avrg_time_to_type_html(-transactions => \@all, -type => 'confirm');
	$html.= CommSecure->_time_last_type_html(-transactions => \@all, -type => 'confirm');
	$html.= CommSecure->_cnt_val_avrg_time_to_type_html(-transactions => \@all);
	$html.= CommSecure->_cnt_val_avrg_time_to_type_html(-transactions => \@all, -type => 'approved');
	$html.= "</table>";
	
	# Which ones are valid and incomplete...
	$html.= "<h2>Incomplete Transactions</h2>";
	if (@incomplete) {
		$html.= "<table border=1 cellspacing=0 cellpadding=0>\n";
		#$html.= CommSecure->_as_html_th();
		$html.= $self->transaction_summary_html_th();
		my $incomplete_value = 0;
		my $count=0;
		foreach (sort {$a->request->date <=> $b->request->date} @incomplete) {
			$html.= $_->transaction_summary_html;
			$count++;
			$incomplete_value+=$_->request->amount;
		}
		$html.= sprintf "<tr><td colspan=2 align=\"right\">Total for $count transactions</td><td align=\"right\" style=\"color: green\">\$%s</td><td colspan=5>-</td></tr>\n", $self->_amt_human($incomplete_value);
		$html.= "</table>\n";
	} else {
		$html.= "<p>No incomplete transactions</p>";
	}

	if (@invalid) {
		# Which ones are invalid
		$html.= "<h2>Invalid Transactions</h2>";
		$html.= "<table border=1 cellspacing=0 cellpadding=0>\n";
		$html.= $self->transaction_summary_html_th();
		foreach (@invalid) {
			$html.= $_->transaction_summary_html;
		}
		$html.= "</table>\n";
	} else {
		$html.= "<p>No invalid transactions</p>";
	}
	return $html;
}

sub _amt_human {
	my $self = shift;
        my $amt = shift;
	my $new_val = "";
	my $cents = sprintf ".%02d", substr($amt, -2);
	$amt = substr($amt, 0, -2)||0;
	while ($amt) {
		$new_val = substr($amt, -3) . $new_val;
		$amt = substr($amt, 0, -3);
		$new_val = "," . $new_val if $amt;
	}
	return $new_val . $cents;
}

sub _time_human {
	my $self = shift;
        my $time = shift;
	return unless $time;
	use integer;
	my $string;
	$string = sprintf "%02d", ($time % 60);
	$time/= 60;
	#return $string unless $time;
	$string = sprintf("%02d", ($time %60)) . ":" . $string;
	$time/=60;
	#return $string unless $time;

	$string = ($time % 24) . ":" . $string;
	$time/=24;
	return $string unless $time;
	$string = ($time % 7) . "d " . $string;
	$time/=7;
	return $string unless $time;
	$string = $time . "w " . $string;
	return $string;
}

sub _as_html_th {
	return "<tr><th>Status</th><th>Cust</th><th>RefNum</th><th>Req'ed</th><th>Amt</th><th>Host</th><th>Val'ed</th><th>Host</th><th>Amt Val</th><th>Comp'ed</th><th>Host</th><th>Amt Res</th><th>Resp</th><th>Auth</th><th>Audit</th><th>Err</th></tr>\n";
}

sub _as_html {
	my $self = shift;
	# Print out this transaction as a table row;
	my $html;
	my ($person, undef) = Person->find(-dbh => $self->dbh, -custnum => $self->request->custnum) if (defined($self->request) && defined($self->request->custnum));
	my $customer_name;
	if (defined($self->request) && defined ($self->request->custnum) && defined($person->name)) {
		$customer_name = $person->name ." (" . $self->request->custnum . ")";
	}elsif(defined($self->request) && defined($self->request->custnum)) {
		$customer_name = $self->request->custnum;
	} else {
		$customer_name = "unknown";
	}
#	defined($self->request) && defined ($self->request->custnum) && ($person->name ." (" . $self->request->custnum . ")") || $self->request->custnum || "-",

	warn "Printing transaction ". $self->refnum;
	warn "Req at " . scalar localtime $self->request->date;
	$html.= sprintf "<tr style=\"background-color: %s\">
			<td>%s</td>
			<td><small>%s</small></td>
			<td><small>%s</small></td>
			<td><small>%s</small></td>
			<td align=\"right\" style=\"color: green\">\$%s</td>
			<td><small>%s</small></td>
			<td><small>%s</small></td>
			<td><small>%s</small></td>
			<td align=\"right\" style=\"color: green\">\$%s</td>
			<td><small>%s</small></td>
			<td><small>%s</small></td>
			<td align=\"right\" style=\"color: green\">\$%s</td>
			<td style=\"color: %s\">%s</td>
			<td><small>%s</small></td>
			<td><small>%s</small></td>
			<td><small>%s</small></td>
			</tr>\n",
			$self->is_approved?"#aaffaa":"#ffaaaa",
			$self->is_valid?"Valid":"INVALID: " . $self->is_invalid,
			$customer_name,
			defined($self->refnum) && $self->refnum || '-',
			defined ($self->request) && defined ($self->request->date) && scalar localtime $self->request->date || '-',
			defined($self->request) && defined($self->request->amount) && $self->_amt_human($self->request->amount) || '-',
			defined($self->request) && defined($self->request->host) && $self->request->host || '-',
			defined($self->validate) && defined($self->validate->[0]->date) && scalar localtime $self->validate->[0]->date || '-',
			defined($self->validate) && defined($self->validate->[0]->host) && $self->validate->[0]->host || '-',
			defined($self->validate) && defined($self->validate->[0]->amount) && $self->_amt_human($self->validate->[0]->amount) || '-',
			defined($self->result) && defined($self->result->[0]->date) && scalar localtime $self->result->[0]->date || '-',
			defined($self->result) && defined($self->result->[0]->host) && $self->result->[0]->host || '-',
			defined($self->result) && defined($self->result->[0]->amount) && $self->_amt_human($self->result->[0]->amount) || '-',
			(defined($self->result) && $self->is_approved)?"green":"red",
			defined($self->result) && defined($self->result->[0]->response) && $self->result->[0]->response || '-',
			defined($self->result) && defined($self->result->[0]->authno) && $self->result->[0]->authno|| '-',
			defined($self->result) && defined($self->result->[0]->audit) && $self->result->[0]->audit || '-',
			defined($self->result) && defined($self->result->[0]->errormsg) && $self->result->[0]->errormsg|| '-',
			;
	return $html;
}

sub _valid_reqs_html {
	my $self = shift;
	my %args = @_;
	return unless defined $args{'-transactions'};
	my $html;
	my $count=0;
	my $value=0;
	foreach (@{$args{'-transactions'}}) {
		#warn "looking at refnum " . $_->refnum;
		next unless defined $_->request;
		$count++;
		$value+=$_->request->amount;
	}
	$html.= sprintf "<tr><td>Total requests</td><td>%s</td></tr>\n", $count;
	$html.= sprintf "<tr><td>Value</td><td align=\"right\" style=\"color: green\">\$%s</td></tr>\n", $self->_amt_human($value);
	return $html;
}

sub _valid_reqs_pending_verify_html {
	my $self = shift;
	my %args = @_;
	return unless defined $args{'-transactions'};
	my $html;
	my $count=0;
	my $value=0;
	foreach (@{$args{'-transactions'}}) {
		next unless defined $_->request;
		next if defined $_->validate;
		$count++;
		$value+=$_->request->amount;
	}
	$html.= sprintf "<tr><td>Total requests pending vrfy</td><td>%s</td></tr>\n", $count;
	$html.= sprintf "<tr><td>Value pending vrfy</td><td align=\"right\" style=\"color: green\">\$%s</td></tr>\n", $self->_amt_human($value);
	return $html;
}

sub _time_last_type_html {
	my $self = shift;
	my %args = @_;
	return unless defined $args{'-transactions'};
	my $html;
	my $time_first;
	my $time_last;
	my @times;
	if (defined $args{'-type'}) {
		if ($args{'-type'} eq "request") {
			my @checked;
			foreach (@{$args{'-transactions'}}) {
				next unless defined $_->request;
				push @checked, $_->request;
			}
			@times = sort {$a->date <=> $b->date} @checked;
		} elsif ($args{'-type'} eq "validate") {
			my @checked;
			foreach (@{$args{'-transactions'}}) {
				next unless $_->validate;
				#warn "It is a " . $_->validate;
				foreach (@{$_->validate}) {
					push @checked, $_;
				}
			}
			@times = sort {$a->date <=> $b->date} @checked;
		} else {
			my @checked;
			foreach (@{$args{'-transactions'}}) {
				next unless defined $_->result;
				foreach (@{$_->result}) {
					push @checked, $_;
				}
			}
			@times = sort {$a->date <=> $b->date} @checked;
		}
	}
	$time_first = $times[0];
	$time_last = pop @times;
	$html.= sprintf "<tr><td>Time of first %s</td><td><small>%s</small></td></tr>\n", $args{'-type'} || 'result', defined($time_first)?scalar(localtime($time_first->date)):"-";
	$html.= sprintf "<tr><td>Time of last %s</td><td><small>%s</small></td></tr>\n", $args{'-type'} || 'result', defined($time_last)?scalar(localtime($time_last->date)):"-";
	return $html;
}

sub _cnt_val_avrg_time_to_type_html {
	my $self = shift;
	my %args = @_;
	return unless defined $args{'-transactions'};
	my $html;
	my $count=0;
	my $time=0;
	my $value=0;
	foreach (@{$args{'-transactions'}}) {
		next unless $_->is_valid;
		if (defined $args{'-type'} && $args{'-type'} eq "validate") {
			next unless $_->validate;
			$count++;
			#warn "Validate is " . ref($_->validate);
			#warn "Request is " . ref($_->request);
			$time+=$_->validate->[0]->date - $_->request->date;
			$value+=$_->validate->[0]->amount;
		} elsif (defined($args{'-type'}) && $args{'-type'} eq "confirm") {
			next unless $_->result;
			$count++;
			$time+=$_->result->[0]->date - $_->validate->[0]->date;
			$value+=$_->result->[0]->amount;
		} elsif (defined($args{'-type'}) && $args{'-type'} eq "approved") {
			next unless $_->result;
			next unless $_->is_approved;
			$count++;
			$time+=$_->result->[0]->date - $_->request->date;
			$value+=$_->result->[0]->amount;
		} else {
			next unless $_->result;
			$count++;
			$time+=$_->result->[0]->date - $_->request->date;
			$value+=$_->result->[0]->amount;
		}

	}
	$html.= sprintf "<tr><td>Number %s</td><td>%s</td></tr>\n", defined $args{'-type'}?$args{-'type'}:"completed", $count;
	$html.= sprintf "<tr><td>Avrg time to %s</td><td>%s</td></tr>\n", defined($args{'-type'})?$args{'-type'}:"complete", $count>0?($self->_time_human($time/$count)):"-";
	$html.= sprintf "<tr><td>Value %s</td><td align=\"right\" style=\"color: green\">\$%s</td></tr>\n", defined($args{'-type'})?$args{'-type'}:"completed", $self->_amt_human($value);
	return $html;
}


sub request {
	my $self = shift;
	if (@_) {
		my $new = shift;
		#warn "Setting requests to " . $new;
		$self->{'request'} = $new;
	}
	return $self->{'request'};
}

sub validate {
	my $self = shift;
	if (@_) {
		#warn "in validate, Came from " . (caller(1))[3];
		my %args = @_;
		$self->{'validate'} = shift;
		push @{$self->{'validate'}}, CommSecure::Validate->new(-dbh => $self->dbh, -refnum => $args{'-refnum'}, -amount => $args{'-amount'}, -host => $args{'-host'}, -time => 'now()');
	}
	return $self->{'validate'};
}

sub result {
	my $self = shift;
	if (@_) {
		$self->{'result'} = shift;
	}
	return $self->{'result'};
}

sub cleared {
	my $self = shift;
	if (@_) {
		my %args = @_;
		my $c = CommSecure::Cleared->new(-dbh => $self->dbh, -refnum => $args{'-refnum'}, -date => time, -amount => $args{'-amount'}, -response => $args{'-response'}, -authno => $args{'-authno'} || '', -audit => $args{'-audit'} || '', -errormsg => $args{'-errormsg'}, -sig => $args{'-sig'}, -host => $args{'-host'});
		$c->save;
		# Update this object
		push @{$self->{'result'}}, $c;
	}
}

sub is_valid {
	my $self = shift;
	if (not $self->request) {
		warn "No request for this transaction refnum " . $self->refnum;
		return;
	}
	if ($self->result and not $self->validate) {
		warn "Result without validate";
		return;
	}

	# We have one request (since we generate this), and zero, one or 
	# more of validates and results. Some of the validates may be bogus, 
	# as may the results. Some may be legitimate.
	
	# SO, what if we had a bogus validate or result before we made the 
	# request for this refnum. A real result can be verified from the 
	# signature, but the validates cannot.

	if (defined($self->request) && defined($self->validate)) {

		#warn "Came from " . (caller(1))[0] . " - " .(caller(1))[3] . " at line " . (caller(1))[2];
		#warn "Req is " . $self->request;

		return unless (defined $self->request->date);
		return unless (defined $self->validate->[0]->date);

		if ($self->request->date > $self->validate->[0]->date) {
			warn "Request before validate for " . $self->refnum;
			return;
		}
		if ($self->request->amount != $self->validate->[0]->amount) {
			warn "Different amounts for " . $self->refnum;
			return;
		}
	}
	if (defined ($self->result) && defined ($self->validate)) {
		return unless defined $self->result->[0]->date;
		return unless defined $self->validate->[0]->date;
		if ($self->validate->[0]->date > $self->result->[0]->date) {
			warn "Validate before result for " . $self->refnum;
			return;
		}
		if ($self->validate->[0]->amount != $self->result->[0]->amount) {
			warn "Different amounts for " . $self->amount;
			return;
		}
	}
	return 1;
}

sub is_invalid {
	my $self = shift;
	return "No request for payment" if (not defined $self->request);
	return "No validate for confirmation" if (defined($self->result) and not defined($self->validate));
	if (defined($self->request) && defined($self->validate)) {
		return "No date for request" unless defined $self->request->date;
		return "No date for validate" unless defined $self->validate->date;
		return "Validate before request" if $self->request->date > $self->validate->date;
		return "Different amounts" if ($self->request->amount != $self->validate->amount);
	}
	if (defined ($self->result) && defined ($self->validate)) {
		return "No date for result" unless defined $self->result->date;
		return "No date for validate" unless defined $self->validate->date;
		return "Result before Validate" if $self->validate->date > $self->result->date;
		return "Different amounts" if ($self->validate->amount != $self->result->amount);
	}
}


sub is_complete {
	my $self = shift;
	return unless $self->is_valid;
	if (defined($self->result)) {
		return 1;
	}
}

sub is_approved {
	my $self = shift;
	return unless $self->is_complete;
	foreach (@{$self->result}) {
		return 1 if ($_->response =~ /^A$/i && $_->verify_sig);
	}
}

sub is_declined {
	my $self = shift;
	return $self->is_complete && $self->result->response =~ /^D$/i && $self->result->verify_sig;
}

sub load_refnum {
	my $self = shift;
	my $refnum = shift;
	#warn "Loading refnum $refnum";
	my @req = CommSecure::Request->find(-dbh => $self->dbh, '-refnum' => $refnum);
	my @val = CommSecure::Validate->find(-dbh => $self->dbh, '-refnum' => $refnum);
	my @res = CommSecure::Cleared->find(-dbh => $self->dbh, '-refnum' => $refnum);
	$self->request($req[0]) if ($#req == 0);
	$self->{'validate'} = \@val if @val;
	$self->result(\@res) if @res;
	return 1;
}

sub find {
	my $self = shift;
	my %args = @_;

	my $dbh = $args{'-dbh'};
	if (defined($dbh) and ref($dbh) !~ "DBI::db") {
		warn "No DBH";
		return;
	}

	# Find the reference numbers and create objects.
	# Ask the request, validate and confirmation objects 
	# for a list of RefNums that match the criteria -- 
	# use the find() method on each. Each find() will give us a 
	# set of objects that match. We then have to merge each set 
	# into transactions...


	my @requests = CommSecure::Request->find(-dbh => $dbh, -refnum => $args{'-refnum'}, -date_min => $args{'-date_request_min'}, -date_max => $args{'-date_request_max'}, -amount_min => $args{'-amount_min'}, -amount_max => $args{'-amount_max'});
	my @validates = CommSecure::Validate->find(-dbh => $dbh, -refnum => $args{'-refnum'}, -date_min => $args{'-date_validate_min'}, -date_max => $args{'-date_validate_max'}, -amount_min => $args{'-amount_min'}, -amount_max => $args{'-amount_max'});
	my @confirms = CommSecure::Cleared->find(-dbh => $dbh, -refnum => $args{'-refnum'}, -date_min => $args{'-date_confirm_min'}, -date_max => $args{'-date_confirm_max'}, -amount_min => $args{'-amount_min'}, -amount_max => $args{'-amount_max'});

	# Get unique (union) refnums
	my %refnums;
	map {$refnums{$_->refnum} = 1} @requests, @validates, @confirms;

	my @transactions;
	foreach (keys(%refnums)) {
		my $transaction = CommSecure->new(-dbh => $args{'-dbh'}, '-refnum' => $_);
		next if (defined $args{'-valid_only'} and not $transaction->is_valid);
		next if (defined $args{'-invalid_only'} and $transaction->is_valid);
		next if (defined $args{'-complete_only'} and not $transaction->is_complete);
		next if (defined $args{'-incomplete_only'} and $transaction->is_complete);
		next if (defined $args{'-approved_only'} and not $transaction->is_approved);
		next if (defined $args{'-declined_only'} and $transaction->is_declined);
		if (defined($args{'-custnum'})) {
			push @transactions, $transaction if ($transaction->custnum eq $args{'-custnum'});
		} else {
			push @transactions, $transaction;
		}
	}
	
	return @transactions;
}


sub start_new_transaction {
	my $self = shift;
	if (@_) {
		my %args = @_;
		# Must have the right arguments
		return unless defined $self->dbh;
		return unless defined $args{'-amount'};
		return unless defined $args{'-host'};
		return unless defined $args{'-merchantnum'};
		return unless defined $args{'-custnum'};
		return unless defined $args{'-description'};
		$self->request(CommSecure->new(-dbh => $self->dbh, -amount => $args{'-amount'}, -host => $args{'-host'}, -merchantnum => $args{'-merchantnum'}, => -custnum => $args{'-custnum'}, -description => $args{'-description'}, -date => now()));
		if (defined $self->request) {
			return $self->request->save;
		} else {
			warn "Couldn't create request";
		}
	}
}

sub validate_transaction {
	my $self = shift;
	if (@_) {
		my %args = @_;
		return unless defined $self->dbh;
		return unless defined $args{'-refnum'};
		return unless defined $args{'-amount'};
		return unless defined $args{'-host'};
		my $v = CommSecure::Validate(-dbh => $self->dbh, -refnum => $args{'-refnum'}, -amount => $args{'-amount'}, -host => $args{'-host'}, -date => now());
		if (defined $v) {
			my $ret = $v->save;
			$self->validate($v) if $ret;
			return $ret;
		}
	}
}

sub confirm_transaction {
	my $self = shift;
	if (@_) {
		my %args = @_;
		return unless defined $self->dbh;
		return unless defined $args{'-refnum'};
		return unless defined $args{'-amount'};
		return unless defined $args{'-response'};
		return unless defined $args{'-host'};
		return unless defined $args{'-sig'};
		my $c = CommSecure::Cleared(-dbh => $self->dbh, -refnum => $args{'-refnum'}, -amount => $args{'-amount'}, -response => $args{'-response'}, -host => $args{'-host'}, -sig => $args{'-sig'}, -date => now(), -authno => $args{'-authno'}, -audit => $args{'-audit'}, -errormsg => $args{'-errormsg'});

		if (defined $c) {
			my $ret = $c->save;
			$self->result($c);
			return $ret;
		}
	}
}

sub transaction_as_html_detail {
	my $self = shift;
	return unless defined $self->refnum;
	my $html = "";
	$html.= sprintf "<table border=1><caption>Request for RefNum %s</caption>", $self->refnum;
	$html.= $self->request->show_as_html_th;
	$html.= $self->request->show_as_html;
	$html.= "</table>";
	if (defined $self->validate)  {
		$html.= "<table border=1><caption>Validate(s)</caption>\n";
		$html.= $self->validate->[0]->show_as_html_th;
		foreach (@{$self->validate}) {
			$html.= $_->show_as_html;
		}
		$html.= "</table>\n";
	}
	if (defined $self->result) {
		$html.= "<table border=1><caption>Result(s)</caption>\n";
		$html.= $self->result->[0]->show_as_html_th;
		foreach(@{$self->result}) {
			$html.= $_->show_as_html;
		}
		$html.= "</table>\n";
	}
	return $html;
}

sub transaction_summary_html_th {
	my $self = shift;
	return "<tr><th>CustNum</th><th>RefNum</th><th>Amt</th><th>Request Date</th><th>Val's</th><th>Res's</th><th><small>Comp</small></th><th><small>Paid</small></th></tr>\n";
}
sub transaction_summary_html {
	my $self = shift;
	use Person;
	my ($person, undef) = Person->find(-dbh => $self->dbh, -custnum => $self->custnum);
	my $customer = "-";
	$customer = sprintf "%s <a href=\"person_show.cgi?custnum=%s\">%s</a>",
		defined($person)?$person->name:"",
		$self->custnum,
		$self->custnum if defined $self->custnum;
	return sprintf "<tr style=\"background-color: %s\"><td><small>%s</small></td><td><small>%s</small></td><td style=\"color: green\" align=\"right\">\$%4.2f</td><td><small>%s</small></td><td align=\"center\">%s</td><td align=\"center\">%s</td><td align=\"center\">%s</td><td align=\"center\">%s</td></tr>\n",
		$self->is_approved?"#aaffaa":"#ffaaaa",
		$customer,
		defined($self->refnum)?"<a href=\"commsecure_showtrans.cgi?refnum=" . $self->refnum . "\">" . $self->refnum . "</a>":"-",
		defined($self->amount)?$self->amount/100:"-0",
		defined($self->request->date)?scalar(localtime($self->request->date)):"-",
		defined($self->validate)?$#{$self->validate} + 1:"-",
		defined($self->result)?$#{$self->result} + 1:"-",
		$self->is_complete?"Y":"N",
		$self->is_approved?"Y":"N";
}

1;
