#! /usr/bin/perl
# $Id: Clearance.pm,v 1.3 2002/08/16 04:26:30 matty Exp $
# Clearance functionality module

package CommSecure::Clearance;
require 5.004;
use Carp 'croak';

use strict;
use vars '$VERSION', '@EXPORT_OK', '@ISA';

$VERSION = qw'$Id: Clearance.pm,v 1.3 2002/08/16 04:26:30 matty Exp $';
@ISA = qw(Exporter);
@EXPORT_OK = qw(hasGoodSig main);

require CGI;
require CGI::Util;
require MIME::Base64;
require Math::BigInt;
require Digest::MD5;

sub main {
	my %fields = form2dict(new CGI);
	checkOriginatedHere($fields{'RefNum'}, $fields{'Amount'});
	if (hasGoodSig(%fields)) {
		recordTransaction(%fields);
	} else {
		recordAttack(%fields);
	}
}

sub hasGoodSig {
	# Verify that the fields supplied contain a valid
	# signature. 'fields' is a dictionary containing
	# keys of: RefNum, Amount, Response,AuthNo,
	# Audit, ErrorMsg and Sig (Some of these fields
	# are optional).

	# Combine and urlencode the parameters (excluding 'Sig').
	# This will produce something like
	# "RefNum=testRef&Amount=440&Response=A&..."
	my (%fields) = @_;
	my (@PossibleParams) = (
		'RefNum', 'Amount', 'Response',
		'AuthNo', 'Audit', 'ErrorMsg');
	my ($param, @params);

	foreach $param (@PossibleParams) {
		if (defined($fields{$param})) {
			my @paramValueParts = map  { CGI::Util::escape($_) } split(' ', $fields{$param});
			my $paramValue = join('+', @paramValueParts);
			push(@params, "$param=$paramValue");
		}
	}
	my $data = join('&', @params);

	print "data $data\n";
	
	# Create an MD5 digest of the URL parameters,
	# and convert it to a number.
	my $digest = Digest::MD5::md5($data);
	my $digestAsNumber = numberFromString($digest);

	print "digestAsNumber $digestAsNumber\n";

	# Decode the (base64-encoded) signature,
	# convert it to a number, and decrypt it with RSA
	# and the CommSecure public key.
	my $sig = MIME::Base64::decode($fields{'Sig'});
	my $sigAsNumber = numberFromString($sig);
	print "sigAsNumber $sigAsNumber\n";

	my ($exponent, $modulus) = getPubKey();
	print "exponent $exponent\n";
	print "modulus $modulus\n";

	# RSA: decrypted = (plain ^ exponent) % modulus,
	# Perl's Math::BigInt's bmodpow() method does this efficiently.
	# Note that calculating (x^y) and _then_ taking the modulus
	# would be _very very_ inefficient.
	my $decryptedSig = $sigAsNumber->bmodpow($exponent, $modulus);
	print "decryptedSig $decryptedSig\n";

	return $decryptedSig == $digestAsNumber;
}

sub getPubKey {
	# Return the exponent and modulus of the CommSecure
	# public key.  Stored as two numbers, one number
	# per line (exponent first) in the file 'CommSecure.pub.key'
	# The CommSecure public key can be fetched from
	# https://clearance.commsecure.com.au/support/pubkey.txt

	open(KEY, 'CommSecure.pub.key');
	my ($exponent, $modulus) = (<KEY>, <KEY>);
	chomp($exponent, $modulus);
	close(KEY);
	$exponent = new Math::BigInt $exponent;
	$modulus = new Math::BigInt $modulus;
	return ($exponent, $modulus);
}

sub form2dict {
	# return simple field->value hash

	my ($cgi) = @_;
	my %fields;
	my $key;
	foreach $key ($cgi->param()) {
		$fields{$key} = scalar($cgi->param($key));
	}
	return %fields;
}

sub checkOriginatedHere {
	# Check that the 'refnum, amount' pair supplied
	# was originally produced by our checkout program.

	my ($refnum, $amount) = @_;
	# Not implemented here.
}

sub recordTransaction {
	my %fields = @_;
	print "Transaction Good\n";
}

sub recordAttack {
	my %fields = @_;
	print "Invalid signature\n";
}

sub numberFromString {
	# Treat each character of 'string' as a byte (in big endian order).
	# Convert to a long and return the converted number.

	my $s = shift;
	my $i = 0;
	my $num = new Math::BigInt 0;
	for ($i = 0; $i < length ($s); $i++) {
		$num = ($num << 8) + ord(substr($s, $i, 1));
	}
	return $num;
}

1;
