package CommSecure::Cleared;
use strict;
use Math::BigInt;
my $VERSION="0.3";
sub Version { $VERSION; }


# Reference CommSecure document "web_protocols"

# A payment confirmation from CommSecure has the following attributes:
# 	RefNum - reference number *we* gave CommSecure
# 	Amount - in cents that was cleared
# 	Response - _a_ccepted or _d_eclient, one char
# OPT	AuthNo - bank authorisation number
# OPT	Audit - bank audit number
# OPT	ErrorMsg - text error message
# 	Sig - Signed digest of the other fields

# In addition to this, we save the following information
# 	Host - that we got the confirmation from
# 	Date - and time we were told this
#

=head1 NAME

CommSecure::Cleared.pm - Modelling CommSecure cleared transactions

=head1 SYNPOSYS

$cleared = CommSecure::Cleared->new(-dbh => $dbh);

=head1 DESCRIPTION

This module permits us to log each 'Cleared' message to the database handle 
that is given to the object during creation. You must have already created 
a valid dtaabase handle and have the correct database tables with permission 
to INSERT into these rows. No other permission should be required (this 
module will only ever keep writing to the table, making a log).

A Transaction has three parts:

=over 3

=item * Request for payment

=item * Validation of reuqest

=item * Result of request in a CLEARED message

=back


A CLEARED message shows the outcome of processing of a Validated Request. 

The common item between each phase of the transaction is the RefNum 
(Reference Number, like a transaction ID). We should also have the same values 
for other attributes of the request, such as Amount, when each of these values 
is present in a phase of the process.

We should also note that we are highly paranoid, and should ensure at every 
stage that the parameters are B<EXACTLY> correct. The final CLEARED message, 
when given to us, should have a cyrptographic Signature in it (in the 
Sig parameter). This library performs the required verification on this 
field.

We are storing all messages that have the required fields, and determine 
their validity in the scope of the transaction, by the values present, and 
by the Signature in this message.

This library is used from the main CommSecure.pm library. There should be 
no requirement to access it directly.

=head1 SEE ALSO

CommSecure.pm

=head1 LICENCE

Cleared.pm - CommSecure Cleared transactions perl library
Copyright (C) 2002 James Bromberger <james@rcpt.to>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=cut


sub new {
	my $proto = shift;
	my $class = ref($proto)||$proto;
	my $self = {};
	bless $self, $class;
	my %args = @_;
	my $ok = 1;
	$self->dbh($args{'-dbh'}) if defined $args{'-dbh'};
	$self->load($args{'-load'}) if defined $args{'-load'};
	if (defined $args{'-date'}) {
		if (not $self->date($args{'-date'})) {
			warn "date bad";
			$ok = 0;
		}
	}
	if (defined $args{'-host'}) {
		if (not $self->host($args{'-host'})) {
			warn "host bad";
			$ok = 0;
		}
	}
	if (defined $args{'-refnum'}) {
		if (not $self->refnum($args{'-refnum'})) {
			warn "refnum bad";
			$ok = 0;
		}
	}
	if (defined $args{'-amount'}) {
		if (not $self->amount($args{'-amount'})) {
			warn "Amount bad";
			$ok = 0;
		}
	}
	if (defined $args{'-response'}) {
		if (not $self->response($args{'-response'})) {
			warn "Response bad";
			$ok = 0;
		}
	}

	$self->authno($args{'-authno'}) if defined $args{'-authno'};
	$self->audit($args{'-audit'}) if defined $args{'-audit'};
	$self->errormsg($args{'-errormsg'}) if defined $args{'-errormsg'};
	if (defined $args{'-sig'}) {
		if (not $self->sig($args{'-sig'})) {
			warn "Bad sig";
			$ok = 0;
		}
	}
	return $self if $ok;
}

sub dbh {
	my $self = shift;
	if (@_) {
		my $new_dbh = shift;
		if (ref($new_dbh) !~ "DBI::db") {
			warn "Not a DBI:db handle: $new_dbh";
			return;
		}
		$self->{'dbh'} = $new_dbh;
	}
	return $self->{'dbh'};
}

sub id {
	my $self = shift;
	if (@_) {
		my $new_id = shift;
		if ($new_id !~ /^\d+$/) {
			warn "Not an ID";
			return;
		}
		$self->{'id'} = $new_id;
	}
	return $self->{'id'};
}

sub date {
	my $self = shift;
	if (@_) {
		my $new_date = shift;
		if (not(defined $new_date) || $new_date !~ /^\d+$/) {
			defined $new_date && warn "Not a date: $new_date";
			defined $new_date || warn "Not a date: (undefined)";
			return;
		}
		$self->{'date'} = $new_date;
	}
	return $self->{'date'};
}
sub date_as_mysql_type {
	my $self = shift;
	return unless $self->date;
	my $d = sprintf "%d/%02d/%02d %02d:%02d:%02d", 
		(localtime($self->date))[5] + 1900,
		(localtime($self->date))[4] + 1,
		(localtime($self->date))[3],
		(localtime($self->date))[2],
		(localtime($self->date))[1],
		(localtime($self->date))[0];
	return $d;
}

sub host {
	my $self = shift;
	if (@_) {
		my $new_host = shift;
		if ($new_host !~ /^[^\s]+$/) {
			#warn "Invalid host: $new_host";
			return;
		}
		$self->{'host'} = $new_host;
	}
	return $self->{'host'};
}

sub refnum {
	my $self = shift;
	if (@_) {
		my $new_ref = shift;
		if ($new_ref !~ /^[\S]+$/) {
			#warn "Invalid RefNo: $new_ref";
			return;
		}
		$self->{'refnum'} = $new_ref;
	}
	return $self->{'refnum'};
}

sub amount {
	my $self = shift;
	if (@_) {
		my $new_amt = shift;
		if ($new_amt !~ /^\d+$/) {
			#warn "Invalid amount: remember, no decimal: $new_amt";
			return;
		}
		$self->{'amount'} = $new_amt;
	}
	return $self->{'amount'};
}


sub response {
	my $self = shift;
	if (@_) {
		my $new_res = shift;
		$self->{'reponse'} = $new_res;
	}
	return $self->{'reponse'};
}


sub transaction_cleared {
	my $self = shift;
	warn "Checking transaction_cleared";
	warn "transaction_cleared: response is ".$self->response;
	return unless defined($self->response);
	return unless $self->verify_sig;
	warn "transaction_cleared: response is ".$self->response;
	return $self->response =~ /^a/i;
}


sub authno {
	my $self = shift;
	if (@_) {
		my $new_auth = shift;
		if (not(defined $new_auth) || length($new_auth) > 4000) {
			defined $new_auth && warn "Invalid AuthNo: $new_auth";
			# This may be undefined since it is an OPTIONAL
			#defined $new_auth || warn "Invalid AuthNo: (undefined)";
			return;
		}
		$self->{'authno'} = $new_auth;
	}
	return $self->{'authno'};
}


sub audit {
	my $self = shift;
	if (@_) {
		my $new_audit = shift;
		if (not(defined $new_audit) || length($new_audit) > 250) {
			defined $new_audit && warn "Invalid Audit: $new_audit";
			# This may be undefined since it is an OPTIONAL
			#defined $new_audit || warn "Invalid Audit: (undefined)";
			return;
		}
		$self->{'audit'} = $new_audit;
	}
	return $self->{'audit'};
}


sub errormsg {
	my $self = shift;
	if (@_) {
		my $new_err = shift;
		# WARN: some error messages can be 270 - 300 long
		if (not(defined $new_err) || length($new_err) > 250) {
			defined $new_err && warn "Invalid error msg: $new_err";
			# This may be undefined since it is an OPTIONAL
			#defined $new_err || warn "Invalid error msg: (undefined)";
			return;
		}
		$new_err ="ERROR: 0(No Error) " if ($new_err eq "ERROR: 0(No Error)");
		$self->{'errormsg'} = $new_err;
	}
	return $self->{'errormsg'};
}


sub sig {
	my $self = shift;
	if (@_) {
		my $new_sig = shift;
		if (length($new_sig) > 2000) {
			warn "Invalid sig: $new_sig";
			return;
		}
		$self->{'sig'} = $new_sig;
	}
	return $self->{'sig'};
}

sub cs_encode {
	my $s = CGI::Util::escape(shift);
	$s =~ s/%20/+/g;
	return $s;
}

sub verify_sig {
	my $self = shift;
	use Digest::MD5 qw(md5);
	use MIME::Base64;
	use Variable;
	my $cs_pub = new Math::BigInt (Variable->new(-dbh => $self->dbh, -name => "CommSecure_PublicKey"));
	my $cs_mod = new Math::BigInt (Variable->new(-dbh => $self->dbh, -name => "CommSecure_Modulus"));
	if (not defined $self->sig) {
		warn "No signature";
		return;
		}

	#my $sig = decode_base64($self->sig);
	my $sig = $self->sig;
	if (not defined $cs_pub) {
		warn "CommSecure public key not loaded";
		return;
	} elsif (not defined $cs_pub) {
		warn "CommSecure modulus not loaded";
		return;
	} elsif (not defined $sig) {
		warn "Signature not available";
		return;
	}

	my @parts;
	push @parts, "RefNum=" . cs_encode($self->refnum) unless ($self->refnum eq "");
	push @parts, "Amount=" . cs_encode($self->amount) unless ($self->amount eq "");
	push @parts, "Response=" . cs_encode($self->response) unless ($self->response eq "");
	push @parts, "AuthNo=" . cs_encode($self->authno) if ($self->authno && u$self->authno ne "");
	push @parts, "Audit=" . cs_encode($self->audit) unless ($self->audit eq "");
	push @parts, "ErrorMsg=" . cs_encode($self->errormsg) unless ($self->errormsg eq "");

	my $data = join "&", @parts;
	#warn "The URL encoded string is $data";
	my $digest = md5($data);
	my $digestAsNumber = _numberFromString($digest);


	my $sigAsNumber = _numberFromString(MIME::Base64::decode($self->sig));
	my $decryptedSig = $sigAsNumber->bmodpow($cs_pub, $cs_mod);
	#warn "digestAsNumber is $digestAsNumber, decryptedSig is $decryptedSig";

	if ($decryptedSig eq $digestAsNumber) {
		#warn "Sig OK";
		return 1;
	}
	warn "Sig Failed";
	return 0;
}

sub _numberFromString {
	# Treat each character of 'string' as a byte (in big endian order).
	# Convert to a long and return the converted number.
	my $s = shift;
	my $i = 0;
	my $num = new Math::BigInt 0;
	for ($i = 0; $i < length ($s); $i++) {
		$num = ($num << 8) + ord(substr($s, $i, 1));
	}
	return $num;
}


sub load {
	my $self = shift;
	my $id = shift;
	if ($id !~ /^\d+$/) {
		warn "Not an ID!";
		return;
	}
	my $sql = "select *, Unix_timestamp(Date) AS ctime_Date from CommSecure_Payments where ID = ?";
	my $sth = $self->dbh->prepare($sql);
	my $res = $sth->execute($id);
	if ($res != 1) {
		warn "Found $res entries";
		return;
		}
	my $ref = $sth->fetchrow_hashref;
	$self->id($ref->{'ID'});
	$self->date($ref->{'ctime_Date'});
	$self->host($ref->{'Host'});
	$self->refnum($ref->{'RefNum'});
	$self->amount($ref->{'Amount'});
	$self->response($ref->{'Response'});
	$self->authno($ref->{'AuthNo'});
	$self->audit($ref->{'Audit'});
	$self->errormsg($ref->{'ErrorMsg'});
	$self->sig($ref->{'Sig'});
	return 1;
}

sub save {
	my $self = shift;
	if (not defined $self->dbh) {
		warn "No DBH to save with";
		return;
	} elsif (not defined $self->host) {
		warn "No host defined";
		return;
	} elsif (not defined $self->refnum) {
		warn "No refnum defined";
		return;
	} elsif (not defined $self->amount) {
		warn "No Amount defined";
		return;
	} elsif (not defined $self->response) {
		warn "No response";
		return;
	} elsif (not defined $self->sig) {
		warn "No Signature";
		return;
	}

	my $sql = "INSERT INTO CommSecure_Payments (Date, Host, RefNum, Amount, Response, Sig";
	$sql.= ", AuthNo" if $self->authno;
	$sql.= ", Audit" if $self->audit;
	$sql.= ", ErrorMsg" if $self->errormsg;
	$sql.= ") VALUES (?, ?, ?, ?, ?, ?";
	$sql.= ", " . $self->dbh->quote($self->authno) if $self->authno;
	$sql.= ", " . $self->dbh->quote($self->audit) if $self->audit;
	$sql.= ", " . $self->dbh->quote($self->errormsg) if $self->errormsg;
	$sql.= ")";

	#warn "SQL to save is $sql";

	my $sth = $self->dbh->prepare($sql);
	my $date = $self->date_as_mysql_type || 'now()';
	#warn "Date is $date";
	my $res = $sth->execute($date , $self->host, $self->refnum, $self->amount, $self->response, $self->sig);
	$sth->finish;
	return $res==1;
}

sub find {
	my $self = shift;
	my %args = @_;
	
	my $dbh = $args{'-dbh'};
	if (not $dbh) {
		warn "No DBH";
		return;
	}

	my $sql = "select ID from CommSecure_Payments ";
	my $where;
	if (defined $args{'-refnum'}) {
		#if ($args{'-refnum'} !~ /^\d+$/) {
		#	warn "Invalid refnum";
		#	return;
		#}
		$where.= " AND " if $where;
		$where.= sprintf "RefNum = %s", $dbh->quote($args{'-refnum'});
	}
	if (defined $args{'-host'}) {
		if (length($args{'-host'}) > 200) {
			warn "Invalid host";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf "Host = %s", $dbh->quote($args{'-host'});
	}
	if (defined $args{'-amount'}) {
		if ($args{'-amount'} !~ /^\d+$/) {
			warn "Invalid amount";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf "Amount = %d", $dbh->quote($args{'-amount'});
	}
	if (defined $args{'-amount_min'}) {
		if ($args{'-amount_min'} !~ /^\d+$/) {
			warn "Invalid amount";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf "Amount >= %d", $dbh->quote($args{'-amount_min'});
	}
	if (defined $args{'-amount_max'}) {
		if ($args{'-amount_max'} !~ /^\d+$/) {
			warn "Invalid amount";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf "Amount <= %d", $dbh->quote($args{'-amount_max'});
	}

	if (defined $args{'-authno'}) {
		if (length($args{'-authno'}) > 200) {
			warn "Invalid Authno";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf "AuthNo = '%s'", $dbh->quote($args{'-authno'});
	}
	if (defined $args{'-errormsg'}) {
		if (length($args{'-errormsg'}) > 200) {
			warn "Invalid ErrorMsg";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf "ErrorMsg = %s", $dbh->quote($args{'-errormsg'});
	}

	if (defined $args{'-date'}) {
		if ($args{'-date'} !~ /^\d+$/) {
			warn "Invalid date";
			return;
		}
		$where .= " AND " if $where;
		$where.= sprintf "Unix_timestamp(Date) = %d", $dbh->quote($args{'-date'});
	}

	if (defined $args{'-date_min'}) {
		if ($args{'-date_min'} !~ /^\d+$/) {
			warn "Invalid date";
			return;
		}
		$where .= " AND " if $where;
		$where.= sprintf "Unix_timestamp(Date) >= %d", $dbh->quote($args{'-date_min'});
	}

	if (defined $args{'-date_max'}) {
		if ($args{'-date_max'} !~ /^\d+$/) {
			warn "Invalid date";
			return;
		}
		$where .= " AND " if $where;
		$where.= sprintf "Unix_timestamp(Date) <= %d", $dbh->quote($args{'-date_max'});
	}

	$sql.= " WHERE " . $where if $where;
	#warn "SQL is $sql";
	my $sth = $dbh->prepare($sql);
	my $res = $sth->execute;
	my @array;
	while (my $ref = $sth->fetchrow_hashref) {
		push @array, CommSecure::Cleared->new(-dbh => $dbh, -load => $ref->{'ID'});
	}
	$sth->finish;
	return @array;
}


sub num_payments {
	my $self = shift;
	my %args = @_;
	my $dbh = $self->dbh || $args{'-dbh'};

	my $sql = "SELECT count(1) as num from CommSecurePayments WHERE Amount != 0";
	my $sth = $dbh->prepare($sql);
	my $res  = $sth->execute;
	if ($res!= 1) {
		$sth->finish;
		warn "Should have got 1 result, got $res";
		return;
	}

	my $ref = $sth->fetchrow_hashref;
	my $num = $ref->{'num'};
	$sth->finish;
	return $num
}

sub total_payments {
	my $self = shift;
	my %args = @_;
	my $dbh = $self->dbh || $args{'-dbh'};

	my $sql = "SELECT sum(Amount) as total from CommSecurePayments WHERE Amount != 0";
	my $sth = $dbh->prepare($sql);
	my $res  = $sth->execute;
	if ($res!= 1) {
		$sth->finish;
		warn "Should have got 1 result, got $res";
		return;
	}

	my $ref = $sth->fetchrow_hashref;
	my $num = $ref->{'total'};
	$sth->finish;
	return $num
}

sub show_as_html_th {
	my $self = shift;
	return "<tr><th>Date</th><th>Host</th><th>RefNum</th><th>Amount</th><th>Response</th><th>AuthNo</th><th>Audit</th><th>ErrorMsg</th><th>Sig</th></tr>\n";

}

sub show_as_html {
	my $self = shift;
	return sprintf "<tr style=\"background-color: %s\"><td><small>%s</small></td><td><small>%s</small></td><td><small>%s</small></td><td style=\"color: green\">%4.2f</td><td>%s</td><td>%s</td><td><small>%s</small></td><td>%s</td><td>%s</td></tr>\n",
		defined($self->transaction_cleared)?"#aaffaa":"#ffaaaa",
		defined($self->sig) && $self->verify_sig && 
		defined($self->date)?scalar(localtime($self->date)):"-",
		defined($self->host)?$self->host:"-",
		defined($self->refnum)?$self->refnum:"-",
		defined($self->amount)?$self->amount/100:"-",
		defined($self->response)?$self->response:"-",
		defined($self->authno)?$self->authno:"-",
		defined($self->audit)?$self->audit:"-",
		defined($self->errormsg)?$self->errormsg:"-",
		defined($self->sig) && $self->verify_sig?"OK":"FAIL";
}

1;
