package CommSecure::Request;
use strict;
use Carp;
my $VERSION="0.3";
sub Version { $VERSION; }


# Reference CommSecure document "web_protocols"

# A request for payment to CommSecure has the following attributes:
#	MerchantNum - out assigned ID
# 	RefNum - reference number *we* gave CommSecure
# 	Amount - in cents that was cleared
#    	CustNum -  16 char personID of some sort
# OPT	MemberNum - 20 char
# OPT	MerchantData - 1024 char
# 	Description - text description

# In addition to this, we save the following information
# 	Host - that we got the confirmation from
# 	Date - and time we were told this
#
# NOTES:
# 	MerchantNum should always be the same, but we save it in case we 
# 	ever have to change, and we then have a complete log anyway that 
# 	is acurate.
#
# 	RefNum: this should not be a uniform serial ID; we dont 
# 		want this easily guessable
# 	CustNum: same - not easily guessable

=head1 NAME

CommSecure::Request.pm - Modelling CommSecure requests for payment

=head1 SYNOPSYS

$request = CommSecure::Request->new(-dbh => $dbh)

=head1 DESCRIPTION

A REQUEST for payment is the start of a tranasaction. The request is what we 
generate when we are about to send a user to CommSecure's server for them 
(the user) to be interragated for their Credit Card details!

We generate a RefNum, and save all parts of the transaction to the table. 
The details of this object are then put into a form (or URL encoded 
link) and the user then submits/follows the link to submit the data to 
CommSecure.

CommSecure then initate an HTTP(S) request to us, and verify that we 
actually did 'refcently' send the user to CommSecure (and we record this 
probe), and then hopefully a few minutes later, COmmSecure initiate another 
HTTP(S) request to us with the results of the processing in a CLEARED message, 
which may contains data indicating it was Accepted or Declined by the 
web clients credit card.

This modules is used by the CommSecure.pm module, and there is no need to 
use it directly.

=head1 SEE ALSO

CommSecure.pm

=head1 LICENCE

Request.pm - CommSecure Cleared transactions perl library
Copyright (C) 2002 James Bromberger <james@rcpt.to>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=cut


sub new {
	
	# This returns a referece to the object if it was successfully 
	# created. It will only fail if the parameters given to it 
	# that it examines are invalid.
	#

	my $proto = shift;
	my $class = ref($proto)||$proto;
	my $self = {};
	bless $self, $class;
	my %args = @_;
	my $ok = 1;
	$self->dbh($args{'-dbh'}) if defined $args{'-dbh'};
	$self->load($args{'-load'}) && return $self if defined $args{'-load'};

	if (defined $args{'-description'}) {
		if (not $self->description($args{'-description'})) {
			warn "descrition bad";
			$ok = 0;
		}
	}
	if (defined $args{'-date'}) {
		if (not $self->date($args{'-date'})) {
			warn "date bad";
			$ok = 0;
		}
	}
	if (defined $args{'-host'}) {
		if (not $self->host($args{'-host'})) {
			warn "host bad";
			$ok = 0;
		}
	}
	if (defined $args{'-refnum'}) {
		if (not $self->refnum($args{'-refnum'})) {
			warn "refnum bad";
			$ok = 0;
		}
	}
	if (defined $args{'-amount'}) {
		if (not $self->amount($args{'-amount'})) {
			warn "amount bad: " . $args{'-amount'} . " from " . (caller(1))[3];
			$ok = 0;
		}
	}
	if (defined $args{'-merchantnum'}) {
		if (not $self->merchantnum($args{'-merchantnum'})) {
			warn "merchant num";
			$ok = 0;
		}
	}
	if (defined $args{'-custnum'}) {
		if (not $self->custnum($args{'-custnum'})) {
			warn "custnum";
			$ok = 0;
		}
	}
	if (defined $args{'-gen-form'}) {
		return if not $ok;
		my $refnum = $self->_generate_refnum();
		return if not defined $refnum;
		#warn "RefNum is $refnum";
		$self->refnum($refnum);
		return unless $self->save;
		return $self->_html_form();

	}

	return $self if $ok;
}

sub dbh {
	#
	# This is the local reference for the database handle
	#

	my $self = shift;
	if (@_) {
		my $new_dbh = shift;
		if (ref($new_dbh) !~ "DBI::db") {
			warn "Not a DBI:db handle: $new_dbh";
			return;
		}
		$self->{'dbh'} = $new_dbh;
	}
	return $self->{'dbh'};
}

sub id {
	#
	# This is the row ID of the record
	#

	my $self = shift;
	if (@_) {
		my $new_id = shift;
		if ($new_id !~ /^\d+$/) {
			warn "Not an ID";
			return;
		}
		$self->{'id'} = $new_id;
	}
	return $self->{'id'};
}

sub date {
	#
	# The date this row was inserted
	#

	my $self = shift;
	if (@_) {
		my $new_date = shift;
		if (not(defined $new_date) || $new_date !~ /^\d+$/) {
			defined $new_date && warn "Not a date: $new_date";
			defined $new_date || warn "Not a date: (undefined)";
			return;
		}
		$self->{'date'} = $new_date;
	}
	return $self->{'date'};
}
sub date_as_mysql_type {
	#
	# This formats the date suitable for feeding to mysql. This is 
	# not needed outside of this module, and not is not considered 
	# to be a 'public' method.
	#

	my $self = shift;
	return unless $self->date;
	my $d = sprintf "%d/%02d/%02d %02d:%02d:%02d", 
		(localtime($self->date))[5] + 1900,
		(localtime($self->date))[4] + 1,
		(localtime($self->date))[3],
		(localtime($self->date))[2],
		(localtime($self->date))[1],
		(localtime($self->date))[0];
	return $d;
}

sub host {
	#
	# This is the host (IP or name) of the remote machine that started 
	# this request to be generated.
	#

	my $self = shift;
	if (@_) {
		my $new_host = shift;
		if ($new_host !~ /^[^\s]+$/) {
			#warn "Invalid host: $new_host";
			return;
		}
		$self->{'host'} = $new_host;
	}
	return $self->{'host'};
}


sub merchantnum {
	#
	# This is a copy of the merchant number; we duplicate this in case 
	# we have to modify our merchant number at any time, and we want to 
	# preserve our history.
	#

	my $self = shift;
	if (@_) {
		my $new_num = shift;
		if ($new_num !~ /^[\w\d_]+$/) {
			#warn "Invalid RefNo: $new_ref";
			return;
		}
		$self->{'merchantnum'} = $new_num;
	}
	return $self->{'merchantnum'};
}

sub refnum {
	#
	# The transaction reference number
	#

	my $self = shift;
	if (@_) {
		my $new_ref = shift;
		if ($new_ref !~ /^[\w\d_\+\/\=]+$/) {
			warn "Invalid RefNo: $new_ref";
			return;
		}
		$self->{'refnum'} = $new_ref;
	}
	return $self->{'refnum'};
}

sub _generate_refnum {
	my $self = shift;
	my $sql = "SELECT ID From CommSecure_Request WHERE RefNum = ?";
	my $sth = $self->dbh->prepare($sql);
	use Digest::MD5 qw(md5_base64);
	my $attempt = md5_base64(rand() . scalar(localtime));
	$attempt =~ s/[\+\&\\\/\s]//g;
	$attempt = substr($attempt, 0, 16);
	my $count = 0;
	while (not($sth->execute($attempt)) && $count < 10) {
		$attempt = md5_base64(rand() . scalar(localtime));
		$attempt =~ s/[\+\&\\\/\s]//g;
		$attempt = substr($attempt, 0, 16);
		$count++;
	}
	return if $count >= 10;
	return $attempt;
}

sub amount {
	#
	# The amount to charge, in cents, with no decimal point
	#

	my $self = shift;
	if (@_) {
		my $new_amt = shift;
		if ($new_amt !~ /^\d+$/) {
			#warn "Invalid amount: remember, no decimal: $new_amt";
			return;
		}
		$self->{'amount'} = $new_amt;
	}
	return $self->{'amount'};
}

sub custnum {
	#
	# The 'customer number', a one way has of our person ID or something
	#

	my $self = shift;
	if (@_) {
		my $new_num = shift;
		if (not(defined($new_num)) || length($new_num)> 16 ) {
			#warn "Invalid custnum: $new_num";
			return;
		}
		$self->{'custnum'} = $new_num;
	}
	return $self->{'custnum'};
}

sub membernum {
	#
	# This seems ot be a repeat of custnum. Not really used...
	#

	my $self = shift;
	if (@_) {
		my $new_num = shift;
		if (not(defined $new_num) || length($new_num) > 20 ) {
			#warn "Invalid membernum: $new_num";
			return;
		}
		$self->{'membernum'} = $new_num;
	}
	return $self->{'membernum'};
}

sub merchantdata {
	# 
	# Any data we want to share with CommSecure about this transaction
	#

	my $self = shift;
	if (@_) {
		my $new_data = shift;
		if (not(defined $new_data) || $new_data !~ /^\d+$/) {
			#warn "Invalid amount: remember, no decimal: $new_amt";
			return;
		}
		$self->{'merchantdata'} = $new_data ;
	}
	return $self->{'merchantdata'};
}

sub description {
	#
	# A human end-customer record to appear on their bill.
	#

	my $self = shift;
	if (@_) {
		my $new_desc = shift;
		if (length($new_desc) > 10000) {
			#warn "Invalid amount: remember, no decimal: $new_amt";
			return;
		}
		$self->{'description'} = $new_desc;
	}
	return $self->{'description'};
}

sub load {
	#
	# Load the required RowID from the database. This requires that there
	# is already a valid database handle (dbh), and that you specify a 
	# rowID to load. This ID is found either by the find method, or 
	# by reference from other objects.
	#

	my $self = shift;
	my $id = shift;
	if ($id !~ /^\d+$/) {
		warn "Not an ID!";
		return;
	}
	my $sql = "select *, Unix_timestamp(Date) AS ctime_Date from CommSecure_Request where ID = ?";
	my $sth = $self->dbh->prepare($sql);
	my $res = $sth->execute($id);
	if ($res != 1) {
		warn "Found $res entries";
		return;
		}
	my $ref = $sth->fetchrow_hashref;
	$self->id($ref->{'ID'});
	$self->date($ref->{'ctime_Date'});
	$self->host($ref->{'Host'});
	$self->refnum($ref->{'RefNum'});
	$self->amount($ref->{'Amount'});
	$self->custnum($ref->{'CustNum'});
	$self->merchantnum($ref->{'MerchantNum'});
	$self->membernum($ref->{'MemberNum'});
	$self->merchantdata($ref->{'MerchantData'});
	$self->description($ref->{'Description'});
	return 1;
}

sub save {
	#
	# This saves the request to the database. It requires that you 
	# already have a valid database handle (dbh), and that the required 
	# attributes of the request are already defined and valid.
	#

	my $self = shift;
	if (not defined $self->dbh) {
		warn "No DBH to save with";
		return;
	} elsif (not defined $self->host) {
		warn "No host defined";
		return;
	} elsif (not defined $self->refnum) {
		warn "No refnum defined";
		return;
	} elsif (not defined $self->amount) {
		warn "No Amount defined";
		return;
	} elsif (not defined $self->description) {
		warn "No Description defined";
		return;
	}

	my $sql = "INSERT INTO CommSecure_Request (Date, Host, RefNum, Amount, Description, CustNum, MerchantNum";
	$sql.= ", MemberNum" if $self->membernum;
	$sql.= ", MerchantData" if $self->merchantdata;
	$sql.= sprintf ") VALUES (%s, ?, ?, ?, ?, ?, ?", $self->date_as_mysql_type || 'now()';

	$sql.= sprintf ", %s",$self->membernum if $self->membernum;
	$sql.= sprintf ", %s",$self->merchantdata if $self->merchantdata;
	$sql.= ")";
	#warn "SQL to save Request is $sql";

	my $sth = $self->dbh->prepare($sql);
	#my $date = $self->date_as_mysql_type || 'now()';
	#warn "Date is $date, sql is $sql";
	my $res = $sth->execute($self->host, $self->refnum, $self->amount, $self->description, $self->custnum, $self->merchantnum);
	$sth->finish;
	return $res==1;
}


sub find {
	# 
	# This will search for the requests that match the supplied criteria, 
	# and return them all as an aray of objects. This prod requires that 
	# you pass it the database handle explicitly.
	#

	my $self = shift;
	my %args = @_;

	my $dbh = $args{'-dbh'};
	if (not $dbh) {
		warn "No dbh";
		return;
	}

	my $sql = "select ID from CommSecure_Request ";
	my $where;
	if (defined $args{'-refnum'}) {
		if ($args{'-refnum'} !~ /^[\S]+$/) {
			carp "Invalid refnum " . $args{'-refnum'};
			return;
		}
		$where .= " AND " if $where;
		$where.= sprintf "RefNum = %s ", $dbh->quote($args{'-refnum'});
	}

	if (defined $args{'-amount'}) {
		if ($args{'-amount'} !~ /^\d+$/) {
			warn "Invalid amount";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf "Amount = %d ", $dbh->quote($args{'-amount'});
	}

	if (defined $args{'-amount_min'}) {
		if ($args{'-amount_min'} !~ /^\d+$/) {
			warn "Invalid amount_min";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf "Amount >= %d ", $dbh->quote($args{'-amount_min'});
	}

	if (defined $args{'-amount_max'}) {
		if ($args{'-amount_max'} !~ /^\d+$/) {
			warn "Invalid amount_max";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf "Amount <= %d ", $dbh->quote($args{'-amount_max'});
	}

	if (defined $args{'-date'}) {
		if ($args{'-date'} !~ /^\d+$/) {
			warn "Invalid date";
			return;
		}
		$where.= " AND " if $where;
		$where .= sprintf " Unix_timestamp(Date) = %d ", $dbh->quote($args{'date'});
	}

	if (defined $args{'-date_min'}) {
		if ($args{'-date_min'} !~ /^\d+$/) {
			warn "Invalid date";
			return;
		}
		$where.= " AND " if $where;
		$where .= sprintf " Unix_timestamp(Date) >= %d ", $dbh->quote($args{'date_min'});
	}

	if (defined $args{'-date_max'}) {
		if ($args{'-date_max'} !~ /^\d+$/) {
			warn "Invalid date";
			return;
		}
		$where.= " AND " if $where;
		$where .= sprintf " Unix_timestamp(Date) <= %d ", $dbh->quote($args{'date_max'});
	}

	if (defined $args{'-custnum'}) {
		if (length($args{'-custnum'}) > 100) {
			warn "Invalid custnum";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf " CustNum = %d ", $dbh->quote($args{'-custnum'});
	}

	$sql.= " WHERE " . $where if $where;

	#warn "SQL for find is $sql";
	my $sth = $args{'-dbh'}->prepare($sql);
	my $res = $sth->execute;
	if ($res < 1) {
		$sth->finish;
		return;
	}
	my @array;
	while (my $ref = $sth->fetchrow_hashref) {
		push @array, CommSecure::Request->new('-dbh' => $args{'-dbh'}, -load => $ref->{'ID'});
	}
	return @array;
}


sub _html_form {
	my $self = shift;
	my %args = @_;
	my $html;
	use CGI;
	my $q = new CGI;
	return unless defined $self->refnum;
	return unless defined $self->amount;
	return unless defined $self->custnum;
	return unless defined $self->description;
	use Variable;
	my $url = Variable->new(-dbh => $self->dbh, -name => "CommSecure_URL");
	my $merchant_num = Variable->new(-dbh=>$self->dbh, -name => "CommSecure_MerchantNum");

	$html.= sprintf "<form method='post' action='%s'>\n", $url;
	$html.= $q->hidden('CustNum', $self->custnum);
	$html.= "\n";
	$html.= $q->hidden('RefNum', $self->refnum);
	$html.= "\n";
	$html.= $q->hidden('Description', 'Linux.conf.au Registration');
	$html.= "\n";
	$html.= $q->hidden('MerchantNum', $merchant_num);
	$html.= "\n";
	$html.= $q->hidden('Amount', $self->amount);
	$html.= "\n";
	$html.= $q->submit(-name => 'Step 1: Go to CommSecure');
	$html.= "\n";
	$html.= "</form>\n";
	return $html;
}


sub show_as_html_th {
	my $self = shift;
	return "<tr><th>Date</th><th>Host</th><th>Merchant</th><th>RefNum</th><th>Amount</th><th>CustNum</th><th>MemberNum</th><th>MerchantData</th><th>Description</th></tr>\n";
}
sub show_as_html {
	my $self = shift;
	return sprintf "<tr><td><small>%s</small></td><td><small>%s</small></td><td>%s</td><td><small>%s</small></td><td style=\"color: green\">%4.2f</td><td><small>%s</small></td><td>%s</td><td>%s</td><td><small>%s</small></td></tr>\n",
		defined($self->date)?scalar(localtime($self->date)):"-",
		defined($self->host)?$self->host:"-",
		defined($self->merchantnum)?$self->merchantnum:"-",
		defined($self->refnum)?$self->refnum:"-",
		defined($self->amount)?$self->amount/100:"-",
		defined($self->custnum)?"<a href=\"person_show.cgi?custnum=" . $self->custnum . "\">" . $self->custnum . "</a>":"-",
		defined($self->membernum)?$self->membernum:"-",
		defined($self->merchantdata)?$self->merchantdata:"-",
		defined($self->description)?$self->description:"-";
}

1;
