package CommSecure::Validate;
use strict;
my $VERSION="0.3";
sub Version { $VERSION; }


# Reference CommSecure document "web_protocols"

# A validate enquiry from CommSecure has the following attributes:
# 	RefNum - reference number *we* gave CommSecure
# 	Amount - in cents that was cleared

# In addition to this, we save the following information
# 	Host - that we got the confirmation from
# 	Date - and time we were told this

=head1 NAME

CommSecure::Validate.pm - Modelling CommSecure Validating Transactions

=head1 SYNOPSYS

$validate = CommSecure::Validate->new(-dbh => $dbh);

=head1 DESCRIPTION

The Validate probe from CommSecure is a response to a user arriving on the 
CommSecure service with parameters in the data submitted to CommSecure 
indicating they have come from your service recently. The response to a 
Validate probe from CommSecure is the text 'Accept' or 'Decline'.

We note that there may be zero, one, or more Validate probes - zero in the case 
the user actually went away, one if everything is going smoothly, or more 
if we have either intruders sending bogus Validate probes, or CommSecure 
sees it necessary to validate twice (perhaps their network into the 
banks failed first time around?). Given this, we must decide on what to 
do if we get multiple validates: there are several choices:

=over 3

=item * Only return Accept to the first one, if it is within 'time' (recently)

=item * Return Accept to every Validate (within 'time')

=item * Return Accept to the first Validate from each unique IP address (within time)

=item * Return Accept to all of the above regarless of the elapsed time

=back


Our objective is to be reliable and reduce unexpected results for the user.


=head1 SEE ALSO

CommSecure.pm

=head1 LICENCE

Validate.pm - CommSecure Validating transactions Perl library
Copyright (C) 2002 James Bromberger <james@rcpt.to>

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

=cut

sub new {
	my $proto = shift;
	my $class = ref($proto)||$proto;
	my $self = {};
	bless $self, $class;
	my %args = @_;
	my $ok = 1;
	$self->dbh($args{'-dbh'}) if defined $args{'-dbh'};
	$self->load($args{'-load'}) if defined $args{'-load'};
	if (defined $args{'-date'}) {
		if (not $self->date($args{'-date'})) {
			warn "date bad";
			$ok = 0;
		}
	}
	if (defined $args{'-host'}) {
		if (not $self->host($args{'-host'})) {
			warn "host bad";
			$ok = 0;
		}
	}
	if (defined $args{'-refnum'}) {
		if (not $self->refnum($args{'-refnum'})) {
			warn "refnum bad";
			$ok = 0;
		}
	}
	if (defined $args{'-amount'}) {
		if (not $self->amount($args{'-amount'})) {
			warn "amount bad";
			$ok = 0;
		}
	}
	if (defined $args{'-save'}) {
		$self->save;
	}

	return $self if $ok;
}

sub dbh {
	my $self = shift;
	if (@_) {
		my $new_dbh = shift;
		if (ref($new_dbh) !~ "DBI::db") {
			warn "Not a DBI:db handle: $new_dbh";
			return;
		}
		$self->{'dbh'} = $new_dbh;
	}
	return $self->{'dbh'};
}

sub id {
	my $self = shift;
	if (@_) {
		my $new_id = shift;
		if ($new_id !~ /^\d+$/) {
			warn "Not an ID";
			return;
		}
		$self->{'id'} = $new_id;
	}
	return $self->{'id'};
}

sub date {
	my $self = shift;
	if (@_) {
		my $new_date = shift;
		if (not(defined $new_date) || $new_date !~ /^\d+$/) {
			defined $new_date && warn "Not a date: $new_date";
			defined $new_date || warn "Not a date: (undefined)";
			return;
		}
		$self->{'date'} = $new_date;
	}
	return $self->{'date'};
}
sub date_as_mysql_type {
	my $self = shift;
	return unless $self->date;
	my $d = sprintf "%d/%02d/%02d %02d:%02d:%02d", 
		(localtime($self->date))[5] + 1900,
		(localtime($self->date))[4] + 1,
		(localtime($self->date))[3],
		(localtime($self->date))[2],
		(localtime($self->date))[1],
		(localtime($self->date))[0];
	return $d;
}

sub host {
	my $self = shift;
	if (@_) {
		my $new_host = shift;
		if ($new_host !~ /^[^\s]+$/) {
			#warn "Invalid host: $new_host";
			return;
		}
		$self->{'host'} = $new_host;
	}
	return $self->{'host'};
}

sub refnum {
	my $self = shift;
	if (@_) {
		my $new_ref = shift;
		if ($new_ref !~ /^[\S]+$/) {
			#warn "Invalid RefNo: $new_ref";
			return;
		}
		$self->{'refnum'} = $new_ref;
	}
	return $self->{'refnum'};
}

sub amount {
	my $self = shift;
	if (@_) {
		my $new_amt = shift;
		if ($new_amt !~ /^\d+$/) {
			#warn "Invalid amount: remember, no decimal: $new_amt";
			return;
		}
		$self->{'amount'} = $new_amt;
	}
	return $self->{'amount'};
}

sub load {
	my $self = shift;
	my $id = shift;
	if ($id !~ /^\d+$/) {
		warn "Not an ID!";
		return;
	}
	my $sql = "select *, Unix_timestamp(Date) AS ctime_Date from CommSecure_Validate where ID = ?";
	my $sth = $self->dbh->prepare($sql);
	my $res = $sth->execute($id);
	if ($res != 1) {
		warn "Found $res entries";
		return;
		}
	my $ref = $sth->fetchrow_hashref;
	$self->id($ref->{'ID'});
	$self->date($ref->{'ctime_Date'});
	$self->host($ref->{'Host'});
	$self->refnum($ref->{'RefNum'});
	$self->amount($ref->{'Amount'});
	return 1;
}

sub save {
	my $self = shift;
	if (not defined $self->dbh) {
		warn "No DBH to save with";
		return;
	} elsif (not defined $self->host) {
		warn "No host defined";
		return;
	} elsif (not defined $self->refnum) {
		warn "No refnum defined";
		return;
	} elsif (not defined $self->amount) {
		warn "No Amount defined";
		return;
	}

	my $sql = sprintf "INSERT INTO CommSecure_Validate (Date, Host, RefNum, Amount) VALUES (%s, ?, ?, ?)", $self->date_as_mysql_type || 'now()';
	#warn "SQL to save is $sql";

	my $sth = $self->dbh->prepare($sql);
	my $date = $self->date_as_mysql_type || 'now()';
	#warn "Date is $date, sql is $sql";
	my $res = $sth->execute($self->host, $self->refnum, $self->amount);
	$sth->finish;
	return $res==1;
}

sub find {
	my $self = shift;
	my %args = @_;

	my $dbh = $args{'-dbh'};
	if (not $dbh) {
		warn "No dbh";
		return;
	}

	my $where;
	if (defined $args{'-refnum'}) {
		$where.= " AND " if $where;
		$where.= sprintf " RefNum = %s ", $dbh->quote($args{'-refnum'});
	}

	if (defined $args{'-date'}) {
		if ($args{'-date'} !~ /^\d+$/) {
			warn "Invalid date";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf " Unix_timestamp(Date) = %d ", $dbh->quote($args{'-date'});
	}
	if (defined $args{'-date_min'}) {
		if ($args{'-date_min'} !~ /^\d+$/) {
			warn "Invalid date min";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf " Unix_timestamp(Date) >= %d ", $dbh->quote($args{'-date_min'});
	}
	if (defined $args{'-date_max'}) {
		if ($args{'-date_max'} !~ /^\d+$/) {
			warn "Invalid date";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf " Unix_timestamp(Date) <= %d ", $dbh->quote($args{'-date_max'});
	}

	if (defined $args{'-amount'}) {
		if ($args{'-amount'} !~ /^\d+$/) {
			warn "Invalid amount";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf " Amount = %d ", $dbh->quote($args{'-amount'});
	}

	if (defined $args{'-amount_min'}) {
		if ($args{'-amount_min'} !~ /^\d+$/) {
			warn "Invalid amount";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf " Amount >= %d ", $dbh->quote($args{'-amount_min'});
	}

	if (defined $args{'-amount_max'}) {
		if ($args{'-amount_max'} !~ /^\d+$/) {
			warn "Invalid amount";
			return;
		}
		$where.= " AND " if $where;
		$where.= sprintf " Amount <= %d ", $dbh->quote($args{'-amount_max'});
	}


	my $sql = "select ID from CommSecure_Validate";
	$sql.= " WHERE " . $where if $where;
	my $sth = $dbh->prepare($sql);
	my $res = $sth->execute;
	if (not ($res > 0)) {
		#warn "None found";
		return;
	}
	my @array;
	while (my $ref = $sth->fetchrow_hashref) {
		push @array, CommSecure::Validate->new('-dbh' => $args{'-dbh'}, -load => $ref->{'ID'});
	}
	return @array;
}


sub show_as_html_th {
	my $self = shift;
	return "<tr><th>Date</th><th>Host</th><th>RefNum</th><th>Amount</th></tr>\n";
}

sub show_as_html {
	my $self = shift;
	return sprintf "<tr><td><small>%s</small></td><td><small>%s</small></td><td><small>%s</small></td><td style=\"color: green\">%4.2f</td></tr>\n", 
		defined($self->date)?scalar(localtime($self->date)):"-",
		defined($self->host)?$self->host:"-",
		defined($self->refnum)?$self->refnum:"-",
		defined($self->amount)?$self->amount/100:"-";
}

1;
