package Invoice;
#
## This script is available under the Gnu General Public Licence (GPL)
##
## Written by James Bromberger (james@rcpt.to).
##
#
use strict;

# An invoice is for an amount to an organisation for one or more 
# items. We define items as one or more registrations (with their 
# options). Thus, then generating an invoice, we must pass the Invoice 
# object one or more Person objects. The invoice then questions each Person 
# for how much they owe (Person->amount_owing()) and then creates a text 
# invoice for the debtor company (a text string). This is then GPG signed, 
# and should be emailed to the responsibel parties.
#
# The invoice is saved, and can be recalled.
#
# When payment is recieved, it is taken as a misc transaction.
#
# Create table Invoice (ID int primary key auto_increment,
# Date datetime,
# Amount float(6,2),
# Text text,
# Debtor text);
# Create table Invoice_People (ID int primary key auto_increment,
# Invoice_ID int,
# CustNum int);

sub new {
	my $proto = shift;
	my $class = ref($proto)||$proto;
	my $self = {};
	bless $self, $class;
	my %args = @_;
	$self->dbh($args{'-dbh'}) if defined $args{'-dbh'};
	$self->load($args{'-load'}) if defined $args{'-load'};
	$self->person($args{'-person'}) if defined $args{'-person'};
	if (defined ($args{'-generate'}) && (defined($args{'-misc'}) || defined($args{'-commsecure'}))) {
		$self->text($self->_generate_text) || return;
		$self->save;
	} elsif (defined $args{'-generate'}) {
		warn "No transaction with generate flag";
		return;
	}
	return $self;
}


sub dbh {
	my $self = shift;
	if (@_) {
		my $new_dbh = shift;
		if (ref $new_dbh !~ "DBI::db") {
			warn "Not a DBI:db handle: $new_dbh";
			return;
		}
		$self->{'dbh'} = $new_dbh;
	}
	return $self->{'dbh'};
}


sub person {
	my $self = shift;
	if (@_) {
		my $new_person = shift;
		if (ref($new_person) ne "Person") {
			warn "Not a person " . ref($new_person);
			return;
		}
		$self->{'person'} = $new_person;
	}
	return $self->{'person'};
}


sub id {
	my $self = shift;
	if (@_) {
		$self->{'id'} = shift;
	}
	return $self->{'id'};
}

sub date {
	my $self = shift;
	if (@_) {
		my $new_date = shift;
		if (not defined $new_date) {
			warn "No date set in Receipt while setting date";
			return;
		}
		if ($new_date !~ /^\d+$/) {
			warn "Not a proper date $new_date!";
			return;
		}
		$self->{'date'} = $new_date;
	}
	return $self->{'date'};
}

sub custnum {
	my $self = shift;
	if (@_) {
		$self->{'custnum'} = shift;
	}
	return $self->{'custnum'};
}

sub text {
	my $self = shift;
	if (@_) {
		my $new_text = shift;
		$self->{'text'} = $new_text;
	}
	return $self->{'text'};
}


sub _generate_text {
	my $self = shift;
	if (not $self->person) {
		warn "No person";
		return;
	}

	my $name = $self->person->name;
	my $amount;
	my $auth_no;
	my $date_paid = scalar localtime;
	if (defined($self->_commsecure) && defined ($self->_commsecure->result)) {
		my $approved;
		# Find the first approved result.
		foreach (@{$self->_commsecure->result}) {
			if ($_->transaction_cleared && $_->amount && $_->date) {
				$approved = $_;
			last;
			}
		}
		# If we didnt find one, return
		if (not $approved) {
			warn "This was not approved";
			return;
		}

		#warn "Got date from cleared message " . ref($approved);
		$date_paid = scalar localtime($approved->date||time);

		# Commsecure deals in cents, we want dollars
		$amount = sprintf "%.2f", ($approved->amount / 100);
		$auth_no = $approved->authno;
		}
	elsif (defined ($self->_misc_transaction) && defined($self->_misc_transaction->date)) {
		warn "Got date from misc transaction";
		$date_paid = scalar localtime $self->_misc_transaction->date;
		$amount = $self->_misc_transaction->amount;
	}
	#warn "Writing reciept for " . $self->person->registration_type->name;
	my $reg_type = sprintf "(%-28s", $self->person->registration_type->name . " registration)";
	my $text =<<FOO;

			     Linux Australia Incorporated


$date_paid WST +0800


Attention: $name

			    	I N V O I C E



INVOICE Linux.conf.au 2003 - The Australian Linux Technical Conference



Thank you for choosing to attend Linux.conf.au 2003. 

Your details are as follows:
FOO
	$text.= sprintf "     Name:\t%s\n", $self->person->name;
	$text.= sprintf "    Email:\t%s\n", $self->person->email_address;
	$text.= sprintf "Telephone:\t%s\n", $self->person->telephone;
	my $addr = $self->person->address;
	$addr =~ s/\n/\n\t\t/m;
	$text.= sprintf "  Address:\t%s\n", $addr;
	$text.= sprintf "\t\t%s\n", $self->person->state;
	$text.= sprintf "\t\t%s %s\n", $self->person->postcode, $self->person->country;
	my $refnum = ("Transaction Reference: " . $self->refnum) || '';
	if (not $refnum) {
		warn "No refnum";
		return;
	}
	my $gst = sprintf "%.2f", $amount * 0.1;
	my $total = sprintf "%.2f", $amount + $gst;
	warn "Doing invoice for $refnum amount $amount type $reg_type";
	$text.= <<FOO;


$refnum

ITEMS						Currency	Amount
---------------------------------------------------------------------------
Linux Australia Membership 			AUD		\\\$   0.00
Linux.conf.au $reg_type	AUD		\\\$ $amount
GST and Taxes					AUD		\\\$   0.00
==============                                                  ===========
Total						AUD		\\\$ $amount



FOO

	$text .=<<FOO;

This incoice is issued by Linux Australia, Incorporated, ABN 56 987 117 479.

Note: The Australian Goods and Services Tax (GST) will NOT be collected 
(Linux Australia is not required to collect GST as it has a turnover under 
AU \\\$100,000 per annum).


Enquiries may be emailed to the organisers: lca-organisers\@conf.linux.org.au.
Details about the conference may be found on the conference web site at 
http://www.linux.conf.au/ until the end of the conference, or at the Linux 
Australia web site at http://www.linux.org.au/.



-- 
 Linux.conf.au: The Australian Linux Techinical Conference
 The University of Western Australia, Perth, January 22 to January 25 2003
 Web: http://www.linux.conf.au/      IRC Help: #lca2003 on irc.openprojects.net

 Linux (R) is a registered trademark of Linus Torvalds
 Linux Australia (TM) is a trademark of Linux Australia Inc. ABN 56 987 117 479
FOO
	return $self->_signed_text($text);
}


sub _signed_text {
	my $self = shift;
	my $clear = shift;
	return unless $clear;
	my $gpg = "echo \"$clear\" | sudo -H -u lca /home/lca/sign-stdin.pl";
	#warn $gpg;
	return join('', `$gpg`);
}


sub load {
	my $self = shift;
	my $id = shift;
	if ($id !~ /^\d+$/) {
		warn "Not an ID: $id";
		return;
	}
	my $sql = "SELECT *, Unix_TIMESTAMP(Date) as UnixDate FROM Receipt WHERE ID = ?";
	my $sth = $self->dbh->prepare($sql);
	my $res = $sth->execute($id);
	if ($res != 1) {
		warn "Didn't get one result";
		return;
	}
	my $ref = $sth->fetchrow_hashref;
	$self->text($ref->{'Text'});
	$self->amount($ref->{'Amount'});
	$self->date($ref->{'UnixDate'});
	$self->id($ref->{'ID'});
	$self->type($ref->{'Type'});

	$sth->finish;
	return 1;
}


sub find {
	my $self = shift;
	my %args = @_;
	my $dbh = $self->dbh || $args{'-dbh'};
	my $sql = "SELECT ID FROM Receipt";
	my $where;
	if (defined $args{'-custnum'}) {
		$where.= " and " if $where;
		$where .= "CustNum = " . $dbh->quote($args{'-custnum'});
	}
	if (defined $args {'-id'}) {
		$where.= " and " if $where;
		$where.= "ID = " . $dbh->quote($args{'-id'});
	}
	$sql .= " WHERE " . $where if defined $where;
	my $sth = $dbh->prepare($sql);
	my $res = $sth->execute;
	if ($res == 0) {
		$sth->finish;
		return;
	}
	#warn "SQL is $sql";
	my @receipts;
	while (my $ref = $sth->fetchrow_hashref) {
		push @receipts, Receipt->new(-dbh => $dbh, -load => $ref->{'ID'});
	}
	return @receipts;
}

sub as_html {
	my $self = shift;
	my $html;
	$html.= sprintf "<p>Receipt number %s for customer %s, issued %s for transaction %s, amount \$%s.</p>\n",
		$self->id, $self->custnum, scalar localtime $self->date,
		$self->refnum, $self->amount;
	$html.= "<pre>\n";
	my $text = $self->text;
	#warn "This text is $text";
	$html.= $text;
	$html.= "</pre>\n";
	return $html;
}

sub list_as_html_th {
	my $self = shift;
	return "<tr><th>Date</th><th>CustNum</th><th>RefNum</th><th>Amount</th></tr>\n";
}
sub list_as_html{
	my $self = shift;
	return sprintf "<tr><td>%s</td><td><small>%s</small></td><td><small>%s</small></td><td><small>%s</small></td><td style=\"color: green\" align=\"right\">%s</td></tr>\n",
		scalar localtime $self->date,
		$self->custnum || "-",
		$self->refnum ||"-",
		$self->amount || "-";
}

sub save {
	my $self = shift;
	my $sql = "INSERT INTO Receipt (CustNum, Text, RefNum, Amount, Type) VALUES (?, ?, ?, ?, ?)";
	my $sth = $self->dbh->prepare($sql);
	my $res = $sth->execute($self->person->custnum, $self->text, $self->refnum, $self->amount, $self->type) or (warn "Could not save reciept: $! " . $self->dbh->errstr && return);
	$sth->finish;
	return ($res == 1);
}

1;
